/*
** scan.c - part of GrabURL 2.0
** Copyright (C) 1996-97 Serge Emond
**
** This program is free software; you can redistribute it and/or
** modify it under the terms of the GNU General Public License
** as published by the Free Software Foundation; either version 2
** of the License, or (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
/* #include <fnmatch.h> */
#include "regex.h"
#include "libgurl.h"
#include "config.h"
#include "macros.h"
#include "config_scan.h"
#include "scan.h"
#include "graburl.h"
#include "uri.h"
#include "log.h"

#ifdef __SASC
#include <dos.h>
#endif

#define MAX_TAG_LEN		1024

int ScanFile(struct Grab *grab)
{
	FILE *fi;
	char tag[MAX_TAG_LEN], *p1, *p2;
	struct Scan_Key *key;
	struct Scan_Ignore *ign=0;
	struct Pat_List *pat;
	int cnt=0, dign, flg, c;
#ifdef __SASC
	char odir[256];
#else
	int odir;
#endif
	struct SURI *suri;

	/* If depth is 0, don't scan! */
	if (grab->uri->depth == 0) {
	  return(0);
	}

#ifdef __SASC
	if (getcwd(odir, 256) == 0) {
#else
	odir = open(".", O_RDONLY);
	if (odir == -1) {
#endif
	  fprintf(stderr, "Error openning current directory\n");
	  exit(EXIT_FAILURE);
	}
	if (chdir(cfg_global.saveroot)) {
	  fprintf(stderr, "ScanFile: Can't CD to SaveRoot (\"%s\")\n",
			  cfg_global.saveroot);
	  exit(EXIT_FAILURE);
	}

	fi = fopen(grab->filename, "r");
#ifdef __SASC
	chdir(odir);
#else
	fchdir(odir); close(odir);
#endif

	if (!fi) {
	  fprintf(stderr, "ScanFile: Can't open file\n");
	  exit(EXIT_FAILURE);
	}

	dign = 0;
	while ((c = fgetc(fi)) != EOF) {
	  if (c == '<') {
		/* A tag... */
		flg = fsgetdelim(fi, tag, MAX_TAG_LEN, 0, ">");
		if (flg == -1) {
		  fprintf(stderr, "ScanFile: error reading file '%s'\n",
				  grab->filename);
		  exit(EXIT_FAILURE);
		}
		if (!flg) {
		  char tok[256];
		  sprintf(tok, "Error scanning \"%s\"\n", grab->filename);
		  putlog(tok, LOG_LIST);
		  break;
		}
		if ((p1 = strchr(tag, '>')) > 0) p1[0] = 0;

		/* Check if we get matching end-ignore */
		if (dign) {
		  p1 = (ign->end ? ign->end : ign->beg);
		  if (!str_nieq(tag, p1, strlen(p1))) {
			dign = 0;
		  }
		}

		/* Check if tag is ignore */

		else {
		  for (ign = scan_cfg.ignore; ign; ign = ign->next) {
			if (!str_nieq(tag, ign->beg, strlen(ign->beg))) {
			  dign = 1;
			  break;
			}
		  }

		  /* Check if contains an url */
		  if (!dign) {
			for (key = scan_cfg.key; key; key = key->next) {
			  p1 = stristr(tag, key->key);
			  if (p1) break;
			}
			
			if (p1 && (p1 = strchr(p1, '='))) {
			  p1++;
			  p2 = ExtractArgument(p1, 0);
			  if (p2) {
				/* An url!!! */
				flg = 1;
				p1 = CompleteUrl(grab->uri->uri, p2);
				if (p1) {
				  /* Check if host matches */
				  if (grab->uri->flags & URI_STAYINHOST) {
					suri = SplitURI(p1);
					if (suri) {
					  if (str_ieq(suri->host, grab->suri->host)) flg = 0;
					  FreeSURI(suri);
					}
				  }
				  
				  /* Check if in include patterns.. */
				  if (flg) for (pat = scan_cfg.accept; pat; pat = pat->next) {
/*					if (fnmatch(pat->pat, p1, FNM_PERIOD) != 0) { */
					if (regexec(&pat->pat, p1, 0, 0, 0) != 0) {
					  flg = 0;
					  break;
					}
				  }
				  
				  /* Check if in exclude patterns ..*/
				  if (flg) for (pat = scan_cfg.skip; pat; pat = pat->next) {
/*					if (fnmatch(pat->pat, p1, FNM_PERIOD) == 0) { */
					if (regexec(&pat->pat, p1, 0, 0, 0) == 0) {
					  flg = 0;
					  break;
					}
				  }
				  
				  /* We assume that depth is either -1 of >0 */
				  if (flg) {
					if (AddURI(grab->uri, p1, (grab->uri->flags & ~URI_TYPES),
							   (grab->uri->depth == -1 ? -1
								: grab->uri->depth -1), 0))
					  cnt++;
				  }
				  free(p1);
				}
				free(p2);
			  }
			}
		  }
		}
	  }
	}
	if (cnt) {
	  char tok[256];
	  sprintf(tok, "ScanFile: added %u urls", cnt);
	  putlog(tok, LOG_LIST);
	}

	fclose(fi);
  	return(cnt);
}

#define CURL_WORK_SIZE 1024

char *CompleteUrl(const char *base, const char *new)
{
	char work[CURL_WORK_SIZE], newp[CURL_WORK_SIZE];
	char *p1, *p2;
	struct SURI *suri;

	/* for now, we drop everything containing '?' */
	/*	if (strchr(new, '?')) return 0; */

	/* We keep it */
	if (stristr(new, "://")) {
	  p1 = (char *)gmalloc(strlen(new)+1);
	  strcpy(p1, new);
	  return p1;
	}
	else if (strchr(new, ':')) {
	  /* Unsupported, ie mailto: */
	  return 0;
	}

	/* 1 - Copy & suri */
	suri = SplitURI(base);
	if (!suri) return(0);

	strcpy(newp, new);
	if (suri->path) strcpy(work, suri->path);
	else work[0]=0;

	/* 2 - check for "#" */
	if (scan_cfg.flags & SCAN_DIESE_SEARCH) {
	  if (new[0]=='#') return(0);

	  p1 = strchr(newp, '#');
	  if (p1) p1[0]=0;
	}

	/* 3 - check for leading '/' */
	if (new[0] == '/') {
	  strcpy(work, suri->protocol);
	  strcat(work, "://");
	  strcat(work, suri->host);
	  if (suri->port) {
		strcat(work, ":");
		sprintf(work + strlen(work), "%u", suri->port);
	  }
	  strcat(work, "/");
	  strcat(work, new+1);
	  strcpy(newp, work);
	}

	/* 4 - process "./" & "../" */
	else {
	  p1 = newp;
	  p2 = strrchr(work, '/');
	  if (p2) p2[0] = 0;
	  while (1) {
		if (p1[0] == '/') {
		  /* skip it */
		  p1++;
		}
		else if (p1[0] == '.') {
		  if (p1[1] == '/' || p1[1]==0) {
			/* ./ - do nothing */
			;
		  }
		  else if (p1[1] == '.' && (p1[2] == '/' || p1[2] == 0)) {
			/* ../ - parent dir */
			p2 = strrchr(work, '/');
			if (!p2) {
			  putlog("Bad URL: we get beyond root dir", LOG_LIST);
			  FreeSURI(suri);
			  return(0);
			}
			p2[0] = 0;
		  }
		  p1 = strchr(p1, '/');
		  if (p1) p1++;
		  else strcat(work, "/");
		}
		else {
		  /* A normal dir.. */
		  p2 = strchr(p1, '/');
		  if (p2) {
			strcat(work, "/");
			strncat(work, p1, p2 - p1);
			p1 = p2+1;
		  }
		  else {
			strcat(work, "/");
			strcat(work, p1);
			p1=0;
		  }
		}
		if ((p1 == 0) || (p1[0] == 0)) break;
	  }

	  strcpy(newp, suri->protocol);
	  strcat(newp, "://");
	  strcat(newp, suri->host);
	  if (suri->port) {
		strcat(newp, ":");
		sprintf(newp + strlen(newp), "%u", suri->port);
	  }
	  if (work[0] != '/') strcat(newp, "/");
	  strcat(newp, work);
	}

	FreeSURI(suri);

	p1 = (char *)gmalloc(strlen(newp) + 1);
	strcpy(p1, newp);

	return p1;
}
